#
# Copyright 2010 Andrey Churin. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import re

from closurebuilder import depstree
from closurebuilder.util import Error
from closurebuilder.manifest import manifest_directive

__all__ = ['Sources', 'Source']

JS_FILE_REGEX = re.compile(r'^.+\.js$')
BASE_REGEX_STRING = '^\s*goog\.%s\(\s*[\'"](.+)[\'"]\s*\)'
PROVIDE_REGEX = re.compile(BASE_REGEX_STRING % 'provide')
REQUIRES_REGEX = re.compile(BASE_REGEX_STRING % 'require')
RESOURCE_REGEX = re.compile(BASE_REGEX_STRING % 'resource')


def get_file_contents(path):
    "Get a file's contents as a string."
    fileobj = open(path)
    try:
        return fileobj.read()
    finally:
        fileobj.close()


class Source(object):
    "Scans a JavaScript source for its provided and required namespaces."

    def __init__(self, filename, strict=False):
        self.provides = set()
        self.requires = set()
        self.resources = []
        self._path = filename
        self._source = get_file_contents(filename)
        self._scan_source()
        if strict:
            if len(self.provides) > 1:
                provides = '  ' + '\n  '.join(self.provides)
                raise Error('Source "%s" provides more than one '
                            'namespace:\n%s\n'%(filename, provides))

    def get_source(self):
        "Gets the source as a string."
        return self._source

    def get_path(self):
        "Returns the path."
        return self._path

    def _scan_source(self):
        "Fill in provides and requires by scanning the source."
        # TODO: Strip source comments first, as these might be in a comment
        # block.  RegExes can be borrowed from other projects.
        source = self.get_source()
        for line in source.splitlines():
            match = PROVIDE_REGEX.match(line)
            if match:
                self.provides.add(match.group(1))
                continue
            match = REQUIRES_REGEX.match(line)
            if match:
                self.requires.add(match.group(1))
                continue
            match = RESOURCE_REGEX.match(line)
            if match:
                self.resources.append(match.group(1))

    def __repr__(self):
        return '%s'%self.get_path()


class Sources(object):
    "Processor for sources.txt file"

    def __init__(self, path):
        self.path = path
        self._tree = None
        self._include_strictly = set()
        self._include_loosely = set()
        self._sources = None
        self._base_js = None

    def _update_includes(self, name, files):
        for i in ('_include_strictly', '_include_loosely'):
            if name != i:
                getattr(self, i).difference_update(files)
        if name:
            getattr(self, name).update(files)

    def include_strictly(self, path):
        path = os.path.abspath(os.path.join(self.path, path))
        if not os.path.isdir(path):
            raise Error('Directory "%s" not found'%path, 8)
        self._update_includes('_include_strictly', self._scan_path(path))

    def include_loosely(self, path):
        path = os.path.abspath(os.path.join(self.path, path))
        if not os.path.isdir(path):
            raise Error('Directory "%s" not found'%path, 8)
        self._update_includes('_include_loosely', self._scan_path(path))
        
    def exclude(self, path):
        path = os.path.abspath(os.path.join(self.path, path))
        if not os.path.isdir(path):
            raise Error('Directory "%s" not found'%path, 8)
        self._update_includes(None, self._scan_path(path))
        
    def include_source_strictly(self, filename):
        filename = os.path.abspath(filename)
        if not os.path.isfile(filename):
            raise Error('File "%s" not found'%filename, 16)
        self._update_includes('_include_strictly', set([filename]))

    def include_source_loosely(self, filename):
        filename = os.path.abspath(filename)
        if not os.path.isfile(filename):
            raise Error('File "%s" not found'%filename, 16)
        self._update_includes('_include_loosely', set([filename]))
        
    def exclude_source(self, filename):
        filename = os.path.abspath(filename)
        if not os.path.isfile(filename):
            raise Error('File "%s" not found'%filename, 16)
        self._update_includes(None, set([filename]))
        
    def _scan_path(self, path):
        result = set()
        for folder, subfolders, files in os.walk(path):
            for filename in files:
                if filename.startswith('.') or not JS_FILE_REGEX.match(filename):
                    continue
                result.add(os.path.abspath(
                              os.path.join(path, folder, filename)))
        return result
        
    def get_sources(self):
        if self._sources is not None:
            return self._sources

        sources = set()
        for filename in self._include_loosely:
            sources.add(Source(filename, strict=False))

        for filename in self._include_strictly:
            sources.add(Source(filename, strict=True))

        self._sources = sources
        return sources

    def get_base_file(self):
        "Returns the base.js file"
        if self._base_js:
            return self._base_js

        def is_closure_base_file(source):
            if os.path.basename(source.get_path()) == 'base.js':
                for line in source.get_source().splitlines():
                    if line.startswith('var goog = goog || {};'):
                        return True
            return False

        base_files = filter(is_closure_base_file, self.get_sources())

        if not base_files:
            raise Error('No closure base.js file found')

        if len(base_files) > 1:
            bases = '  ' + '\n  '.join(map(lambda x: x.get_path(), base_files))
            raise Error('More than one Closure base.js files found at '
                             'these paths:\n%s' % bases)

        self._base_js = base_files[0]
        return self._base_js

    def get_dependencies(self, namespace, exclude = []):
        "Get source dependencies, in order, for the given namespaces"
        if not self._tree:
            self._tree = depstree.DepsTree(self.get_sources())

        return self._tree.get_dependencies(namespace, exclude)

    def find_sources(self, provide=None, require=None):
        "Finds file by the specified namespace"
        result = []
        for i in self.get_sources():
            if provide:
                if provide in i.provides:
                    return i
            elif require:
                if require in i.requires:
                    result.append(i)
        return result

    @manifest_directive
    def scan_(self, opt, value):
        # deprecated, use include strictly instead    
        "scan %"
        self.include_strictly(value)

    @manifest_directive
    def scan_loose_(self, opt, value):
        # deprecated, use include instead
        "scan loose %"
        self.include_loosely(value)
        
    @manifest_directive
    def include_(self, opt, value):
        "include %"
        self.include_loosely(value)

    @manifest_directive
    def include_loosely_(self, opt, value):
        "include % loosely"
        self.include_loosely(value)

    @manifest_directive
    def include_strictly_(self, opt, value):
        "include % strictly"
        self.include_strictly(value)        
                
    @manifest_directive
    def exclude_(self, opt, value):
        "exclude %"
        self.exclude(value)

    @manifest_directive
    def include_source_(self, opt, value):
        "include source %"
        self.include_source_loosely(value)

    @manifest_directive
    def include_source_loosely_(self, opt, value):
        "include source % loosely"
        self.include_source_loosely(value)
        
    @manifest_directive
    def include_source_strictly_(self, opt, value):
        "include source % strictly"
        self.include_source_strictly(value)
        
    @manifest_directive
    def exclude_source_(self, opt, value):
        "exclude source %"
        self.exclude_source(value)
